package de.renew.gui;

import java.util.*;
import java.awt.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.framework.*;
import de.renew.simulator.*;

import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.StreamCorruptedException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;
import de.renew.util.StringUtil;

public class CPNInstanceDrawing extends StandardDrawing
                                implements DrawingChangeListener,
					   DrawingContext {

  public static String ID_MACRO="$ID";

  private static Hashtable drawingsByInstance = new Hashtable();

  private NetInstance netInstance;
  private CPNDrawing cpnDrawing;
  private RenewMode mode;
  private Hashtable instanceLookup;

  private CPNInstanceDrawing(RenewMode mode, 
			     NetInstance netInstance, 
			     CPNDrawing drawing)
  {
    this.mode = mode;
    this.cpnDrawing = drawing;
    connect(netInstance);
    cpnDrawing.addDrawingChangeListener(this);
  }

  RenewMode getMode() {
      return mode;
  }

  /** Get the NetInstance associated with this drawing. */
  public NetInstance getNetInstance() {
      return netInstance;
  }
    
  public Rectangle displayBox() {
    return super.displayBox().union(cpnDrawing.displayBox());
  }

  /**
   * @param mode needed to look up the responsible simulation
   * @param netInstance to look up the drawing for
   * @return CPNInstanceDrawing which displays the given netInstance.
   *         If there is currently no such drawing, one is created.
   *         If there is not even any CPNDrawing matching the net
   *         known to the active simulation, returns 
   *         <code>null</code>.
   **/
  static CPNInstanceDrawing getInstanceDrawing(RenewMode mode,
					       NetInstance netInstance)
  {
    // forbid concurrent searches for instance drawings.
    synchronized (drawingsByInstance) {
      // find instance drawing:
      if (drawingsByInstance.containsKey(netInstance))
	return (CPNInstanceDrawing)drawingsByInstance.get(netInstance);
      else {
	// find drawing to base the instance drawing on
	CPNSimulation sim = mode.getSimulation();
	CPNDrawing drawing = sim.getDrawing(netInstance.getNet().getName());
	if (drawing != null) 
	  return new CPNInstanceDrawing(mode, netInstance, drawing);
	else 
	  System.err.println("To display a net instance, an open"
			     +" drawing of the instantiated net is"
			     +" needed. No such drawing found for"
			     +" net \""+netInstance.getNet().getName()
			     +"\".");
	  return null;
      }
    }
  }

  /** 
   * Returns all instance drawings based on the specified CPNDrawing.
   * Returns <code>null</code>, if no such drawing exists.
   **/
  public static Enumeration getDependentInstanceDrawings(CPNDrawing baseDrawing) {
    Enumeration allDrawings = drawingsByInstance.elements();
    Vector dependentDrawings = new Vector();
    while (allDrawings.hasMoreElements()) {
      CPNInstanceDrawing oneDrawing = (CPNInstanceDrawing)allDrawings.nextElement();
      if (oneDrawing.cpnDrawing == baseDrawing)
	dependentDrawings.addElement(oneDrawing);
    }

    if (dependentDrawings.size() > 0)
      return dependentDrawings.elements();
    else
      return null;
  }
  
  public void draw(Graphics g) {
      //g.setColor(Color.yellow.brighter());
      //Rectangle b=displayBox();
      //g.fillRect(b.x,b.y,b.width,b.height);
    cpnDrawing.draw(g,this);
    super.draw(g);
  }

  public InstanceFigure getInstanceFigure(Figure fig) {
      return (InstanceFigure)instanceLookup.get(fig);
  }

  public boolean isHighlighted(Figure fig) {
      FigureWithHighlight node=cpnDrawing.getFigureForHighlight(fig);
      if (node != null)
	  fig = node;
      InstanceFigure instFig=getInstanceFigure(fig);
      if (instFig!=null) {
	  return instFig.isHighlighted();
      }
      return false;
  }

  public boolean isVisible(Figure fig) {
      return !(((fig instanceof CPNTextFigure
		 && ((CPNTextFigure)fig).getType() == CPNTextFigure.INSCRIPTION
		 && ((CPNTextFigure)fig).parent() instanceof PlaceFigure))
	       || fig == cpnDrawing.getIconFigure());
  }

  public String expandMacro(String text) {
      return expandMacro(text,netInstance);
  }

  public static String expandMacro(String text, NetInstance instance) {
      return StringUtil.replace(text,ID_MACRO,instance.getID());
  }

  public void connect(NetInstance netInstance) {
    if (netInstance != null) {
      if (this.netInstance != null)
	drawingsByInstance.remove(this.netInstance);
      this.netInstance = netInstance;
      drawingsByInstance.put(netInstance,this);
      setName(netInstance.toString());
      instanceLookup = new Hashtable();
      // FigureLookup figureLookup = mode.getSimulation().getFigureLookup();
      // Now we build us our own lookup...
      FigureLookup figureLookup = buildFigureLookup(netInstance);
      FigureEnumeration figures = cpnDrawing.figures();
      while (figures.hasMoreElements()) {
	Figure figure = figures.nextFigure();
	if (figure instanceof PlaceFigure) {
	  PlaceFigure placeFig = (PlaceFigure)figure;
	  Place place;
          if (placeFig instanceof VirtualPlaceFigure) {
             place=figureLookup.get( ((VirtualPlaceFigure)placeFig).getPlace() );
          } else {
             place=figureLookup.get(placeFig);
          }
	  if (place!=null) {
	    Figure pif=new PlaceInstanceFigure(this,placeFig,netInstance.getInstance(place));
	    instanceLookup.put(placeFig,pif);
	    add(pif);
	    sendToBack(pif);
	  }
	} else if (figure instanceof TransitionFigure) {
	  TransitionFigure transitionFig = (TransitionFigure)figure;
	  Transition transition=figureLookup.get(transitionFig);
	  if (transition!=null) {
            Figure tif=new TransitionInstanceFigure(this,transitionFig,
						    netInstance.getInstance(transition));
	    instanceLookup.put(transitionFig,tif);
	    add(tif);
	  }
	}
      }
    }
  }


  /**
   * Tries to map graphical figures of an open CPNDrawing
   * to net elements of the given NetInstance.
   * @param instance NetInstance to try to map
   * @return FigureLookup (which is empty if the mapping was unsuccessfull)
   **/
  private FigureLookup buildFigureLookup(NetInstance netInstance) {

    Hashtable idMapping = new Hashtable();
    Hashtable figureMapping = null;
    FigureLookup figureLookup;

    Net net = netInstance.getNet();
    String netName = net.getName();
    CPNDrawing netDrawing = mode.getSimulation().getDrawing(netName);
    if (netDrawing == null) {
      System.err.println("No open drawing found matching the "
			 +" compiled net \""+netName+"\"");
    } else {
      //System.out.println("Mapping graphical to compiled "
      //                  +"elements for net \""+netName
      //                  +"\"...");
      Enumeration  netElements;
      Place        currentPlace;
      Transition   currentTransition;
      FigureWithID currentFigure;
      int          currentID;
      boolean warning1Found, warning2Found, warning3Found;
      boolean errorFound;

      // Proceeding: First build a mapping from IDs to
      // net elements, then try to map net figures
      // to these elements using their ID.

      figureMapping = new Hashtable();

      // Fill in all places into the ID mapping
      netElements = net.places();
      while (netElements.hasMoreElements()) {
	currentPlace = (Place)netElements.nextElement();
	currentID = currentPlace.getID();
	idMapping.put(new Integer(currentID), currentPlace);
      }

      // Fill in all transitions into the ID mapping
      netElements = net.transitions();
      while (netElements.hasMoreElements()) {
	currentTransition = (Transition)netElements.nextElement();
	currentID = currentTransition.getID();
	idMapping.put(new Integer(currentID), currentTransition);
      }

      // Try to map figures to net elements by ID
      // There are 4 problem categories:
      //
      // 1 Missing figure for a net element
      //   (e.g. the user added elements to the drawing)
      // 2 Missing net element for a figure
      //   (e.g. the user removed elements from the drawing)
      // 3 Different types of elements with the same ID
      //   (e.g. the user removed and added elements)
      // 4 Same element type, but modified attributes:
      //   Such errors are neither recognized nor handled.
      //
      // If one of the problems 1-3 occurs, it is
      // impossible to map the offending net element.
      // But as other elements may still be mapable,
      // a warning message seems to be appropriate.
      warning1Found = false;
      warning2Found = false;
      warning3Found = false;
      errorFound = false;

      // Check for problem type 1:
      // Get all IDs of the ID mapping and check if
      // corresponding figures exist.
      // This check is done in an extra pass because it needs
      // the inverse relation.
      netElements = idMapping.keys();
      while (!errorFound && netElements.hasMoreElements()) {
	currentID = ((Integer)netElements.nextElement()).intValue();
	if (netDrawing.getFigureWithID(currentID) == null) {
	  //System.err.println("Missing figure: "+currentID);		
	  warning1Found = true;
	}
      }

      // Iterate through all figures and look up the
      // corresponding (with same ID) net elements.
      netElements = netDrawing.figures();
      while (!errorFound && netElements.hasMoreElements()) {
	// If there are figures without ID support,
	// they are not relevant, skip them.
	try {
	  currentFigure = (FigureWithID)netElements.nextElement();
	} catch (ClassCastException e) {
	  currentFigure = null;
	}
	if (currentFigure != null) {
	  currentID = currentFigure.getID();
	  // Check only places and transitions as only their
	  // IDs are in the idMapping.
	  // As there is no common superclass supporting IDs,
	  // the code has to be written twice...
	  if (currentFigure instanceof PlaceFigure) {
	    try {
	      currentPlace = (Place)idMapping.get(new Integer(currentID));
	      if (currentPlace != null) {
		figureMapping.put(currentFigure, currentPlace);
	      } else {
		//System.err.println("Missing net element: "+currentID);		
		warning2Found = true;
	      }
	    } catch (ClassCastException e) {
	      //System.err.println("Mismatch: "+currentID);		
	      warning3Found = true;
	    }
	  } else if (currentFigure instanceof TransitionFigure) {
	    try {
	      currentTransition = (Transition)idMapping.get(new Integer(currentID));
	      if (currentTransition != null) {
		figureMapping.put(currentFigure, currentTransition);
	      } else {
		//System.err.println("Missing net element: "+currentID);		
		warning2Found = true;
	      }
	    } catch (ClassCastException e) {
	      //System.err.println("Mismatch: "+currentID);		
	      warning3Found = true;
	    }
	  }
	}
      }

      // Print the appropriate warning messages
      if (warning1Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" does not contain one or more net"
			   +" elements found in the compiled net.");
	System.err.println("Some logical elements will not be"
			   +" displayed.");
      }
      if (warning2Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" contains graphical net elements"
			   +" not found in the compiled net.");
	System.err.println("Some graphical elements will not be"
			   +" functional.");
      }
      if (warning3Found) {
	System.err.println("Warning: Drawing \""+netName+"\""
			   +" contains graphical elements which map"
			   +" to compiled net elements of incompatible"
			   +" type.");
	System.err.println("Some graphical elements will not be"
			   +" functional.");
      } 

      // If some fatal problem was encountered,
      // deliver no mapping instead of an invalid
      // one.
      if (errorFound) {
	figureMapping = null;
      }
    }

    // Finally convert the new built mapping
    // into a figureLookup
    figureLookup = new FigureLookup(figureMapping);
    return figureLookup;
  }

  public void release() {
    cpnDrawing.removeDrawingChangeListener(this);
    if (netInstance != null) {
      drawingsByInstance.remove(netInstance);
      //System.err.println("Instance Drawing removed from list.");
      instanceLookup=null;
      netInstance = null;
    }
    super.release();
  }


  /**
   *  Sent when an area is invalid
   */
  public void drawingInvalidated(DrawingChangeEvent e) {
    // System.out.println("Source invalidated: "+e.getInvalidatedRectangle());
    DrawingChangeEvent myEvent = new DrawingChangeEvent(this,
      e.getInvalidatedRectangle());
    Enumeration listeners = drawingChangeListeners();
    while (listeners.hasMoreElements()) {
      DrawingChangeListener listener = (DrawingChangeListener)listeners.nextElement();
      listener.drawingInvalidated(myEvent);
      listener.drawingRequestUpdate(myEvent);
    }
  }

  /**
   *  Sent when the drawing wants to be refreshed
   */
  public void drawingRequestUpdate(DrawingChangeEvent e) {
    //System.out.println("Source requests update: "+e.getInvalidatedRectangle());
  }


   /**
     * Returns whether drawing has been modified since last save.
     */
    public boolean isModified() {
       return false;  // an InstanceDrawing should not be saved!
    }

  /**
   * Writes all currently known <code>NetInstances</code>s
   * to the given stream. The written information is <b>not</b>
   * guaranteed to describe the complete current simulation
   * state!
   * A <code>NetInstances</code> is known, if it is displayed
   * by a <code>CPNInstanceDrawing</code>.
   * <p>
   * If the given <code>ObjectOutput</code> is a <code><b>
   * de.renew.util.RenewObjectOutputStream</b></code>, its
   * feature of cutting down the recursion depth by delaying
   * the serialization of some fields will be used.
   * </p><p>
   * <b>Side effect:</b> Any open <code>BindingSelectionFrame</code>
   * will be closed on execution of this method!
   * <br></br>
   * <b>Caution:</b> In order to get consistent data written
   * to the stream, you have to ensure that there are no
   * concurrent modifications of the simulation state.
   * This method is not able to lock the simulation.
   * </p>
   *
   * Added Feb 29 2000  Michael Duvigneau
   *
   * @param output target stream (see note about RenewObjectOutputStream)
   * @see de.renew.util.RenewObjectOutputStream
   **/
  public static void saveAllInstances(ObjectOutput output) 
         throws IOException {

    Enumeration saveList;
    RenewObjectOutputStream rOut = null;
    if (output instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)output;

    if (rOut != null) rOut.beginDomain(CPNInstanceDrawing.class);

    // Close the BindingSelectionFrame (otherwise it would
    // be stored with its transition).
    TransitionInstanceFigure.closeBindingSelectionFrame();

    // Save all currently known instances.
    output.writeInt(drawingsByInstance.size());
    saveList=drawingsByInstance.keys();
    while (saveList.hasMoreElements()) {
      output.writeObject(saveList.nextElement());
    }

    // If a RenewObjectOutputStream is used, write
    // all delayed fields NOW. 
    if (rOut != null) rOut.writeDelayedObjects();

    if (rOut != null) rOut.endDomain(CPNInstanceDrawing.class);
  }

  /**
   * Reads NetInstances stored by <code>saveAllInstances()
   * </code>.
   * <p>
   * This method does neither automatically add the new<code>
   * NetInstance</code>s to the list of known instances nor
   * open a CPNInstanceDrawing for them.
   * The reason is that first the active simulation has to
   * rebuild its FigureLookup.
   * It is then left up to the caller to request the
   * appropriate <code>CPNInstanceDrawing</code>s via
   * <code>getInstanceDrawing()</code>.
   * </p><p>
   * If the given <code>ObjectInput</code> is a <code>
   * de.renew.util.RenewObjectInputStream</code>, the
   * neccessary steps to cover delayed serialization will
   * be made.
   * </p><p>
   * The ObjectInputStream will be read using
   * <code>de.renew.util.ClassSource</code> to provide
   * its ability of reloading all user defined classes.
   * </p>
   *
   * Added Feb 29 2000  Michael Duvigneau
   *
   * @return Enumeration of NetInstances
   * @param input source stream (see note about RenewObjectInputStream)
   * @see #saveAllInstances
   * @see de.renew.util.ClassSource
   * @see de.renew.util.RenewObjectInputStream
   **/
  public static Enumeration loadInstances(ObjectInput input) 
         throws IOException, ClassNotFoundException {

    // Read all stored NetInstances.
    int count = input.readInt();
    Vector newInstances = new Vector(count);
    try {
      for (int i = 0; i < count; i++) {
	newInstances.addElement(
	  (NetInstance)de.renew.util.ClassSource.readObject(input));
      }      
    } catch (ClassCastException e) {
      throw new StreamCorruptedException (
		"Object other than NetInstance found "
		+"when looking for net instances: "
		+e);
    }
    
    // If a RenewObjectInputStream is used, read
    // all delayed fields NOW.
    if (input instanceof RenewObjectInputStream)
      ((RenewObjectInputStream)input).readDelayedObjects();

    // Return the list of read NetInstances.
    return newInstances.elements();
  }
}
