package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.function.Executor;
import de.renew.util.Value;
import de.renew.util.Null;
import java.lang.reflect.Array;

class FlexibleArcBinder implements Binder {
  Variable variable;
  TokenCollector collector;
  FlexibleArc arc;

  FlexibleArcBinder(Variable variable, TokenCollector collector,
    FlexibleArc arc)
  {
    this.variable=variable;
    this.collector=collector;
    this.arc=arc;
  }

  public int bindingBadness(Searcher searcher) {
    // The possibly incomplete value in the variable
    // will not be stored anywhere, so that it is
    // not required to make a copy of it.
    if (Unify.isBound(variable)) {
      return 1;
    } else {
      // We must not try to bind the variable.
      return BindingBadness.max;
    }
  }

  public void bind(Searcher searcher) {
    // Make sure that the place instance notifies
    // the searchable if its marking changes,
    // because in that case the possible binding would have
    // to be rechecked. This is only required for input
    // arcs, because output arcs cannot disable or enable
    // a transition.
    collector.registerAt(searcher);

    if (Unify.isBound(variable)) {
      // The array is analysed locally and need not be copied.
      Object tokens=variable.getValue();

      if (tokens==null) {
	// Let's ignore this arc. Probably the null value is meant as
	// a substitute for an empty array.
	searcher.search();
      } else if (tokens.getClass().isArray()) {
	Class elementType=tokens.getClass().getComponentType();

	LinkedList removedTokens=new LinkedList();

	int n=Array.getLength(tokens);

	boolean success=true;
	for (int i=0;success && i<n;i++) {
	  Object tok=null;
	  try {
	    // Extract a single token and copy it, so that
	    // backtracking does not hurt.
	    tok=Unify.copyBoundValue(Array.get(tokens,i));
	    
	    if (elementType.isPrimitive()) {
	      tok=new Value(tok);
	    } else {
	      // Is this really neccessary? The array should not
	      // contain any objects that are subject to backtracking.
	      // But let's be safe.
	      tok=Unify.copyBoundValue(tok);
	    }

	    if (arc.forwardFunction!=null) {
	      Object convertedToken=arc.forwardFunction.function(tok);
	      
	      if (arc.backwardFunction!=null) {
		Object restoredToken=
		  arc.backwardFunction.function(convertedToken);
		Unify.unify(tok,restoredToken,null);
	      }

	      tok=convertedToken;
	    }
   
	    success=collector.collectToken(tok);
	  }
	  catch (Exception e) {
	    e.printStackTrace();
	    success=false;
	  }
	  
	  if (success) {
	    removedTokens.insertFirst(Null.convert(tok));
	  }
	}

	if (success) {
	  // All reservations succeeded.
	  searcher.search();
	}
	  
	// Undo the successful reservations.
	CollectionEnumeration enum=removedTokens.elements();
	while (enum.hasMoreElements()) {
	  collector.freeToken(Null.convert(enum.nextElement()));
	}
      }
    } else {
      // This should not happen.
      throw new RuntimeException("Flexible arc binder was invoked for an "+
	"incomplete value.");
    }
  }
}
