package de.renew.simulator;

import de.renew.unify.Impossible;
import collections.*;
import java.util.Hashtable;
import java.util.Enumeration;

import java.io.Serializable;
import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.StreamCorruptedException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

public class Net implements Serializable {

  static final long serialVersionUID = 4424943946669800287L;

  private static Hashtable netsByName=new Hashtable();
  public static Net forName(String name) {
    Net net=(Net)((Hashtable)netsByName).get(name);
    if (net==null) {
      System.err.println("Net "+name+" not loaded.");
    }
    return net;
  }
  public static void forgetAllNets() {
    netsByName.clear();
  }

  UpdatableSet places=new HashedSet();
  UpdatableSet transitions=new HashedSet();

  String name=null;

  boolean earlyTokens=false;

  private int netCount=0;

  public Net() {
    name=null;
  }

  public Net(String name) {
    this.name=name;

    if (name!=null) {
      netsByName.put(name,this);
    }
  }

  public CollectionEnumeration places() {
    return places.elements();
  }

  public CollectionEnumeration transitions() {
    return transitions.elements();
  }

  int makeNetNumber() {
    return ++netCount;
  }

  public String toString() {
    return name;
  }

  public String getName() {
    return name;
  }

  NetInstance makeInstance() throws Impossible {
    // Instead of creating a generic NetInstance, we might also
    // want to create a specialized NetInstance that supports
    // own methods and implements interfaces.
    return new NetInstanceImpl(this);
  }

  // Unlike the previous method, this publicly accessible
  // method immediately confirms the creation of the
  // generated net instance, so that external users do not need
  // to know about confirmation issues.
  public NetInstance buildInstance() {
    try {
      NetInstance instance=makeInstance();
      instance.createConfirmation(SilentTracer.INSTANCE);
      return instance;
    } catch (Impossible e) {
      throw new RuntimeException("Could not make net instance.");
    }
  }

  void add(Place place) {
    places.include(place);
  }
  void add(Transition transition) {
    transitions.include(transition);
  }
  void remove(Place place) {
    places.exclude(place);
  }
  void remove(Transition transition) {
    transitions.exclude(transition);
  }

  public void setEarlyTokens(boolean earlyTokens) {
    this.earlyTokens=earlyTokens;
  }



  /**
   * Serialization method, behaves like default writeObject
   * method except using the domain trace feature, if the
   * output is a RenewObjectOutputStream.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) rOut.endDomain(this);
  }

  /**
   * Deserialization method, behaves like default readObject
   * method and makes the net known to <code>Net.forName()</code>.
   **/
  private void readObject(java.io.ObjectInputStream in)
      throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    if (name != null) {
      netsByName.put(name, this);
    }
  }

  /**
   * Writes all currently known <code>Net</code>s
   * to the given stream. The written information 
   * describes the static part of all compiled nets.
   * <p>
   * If the given <code>ObjectOutput</code> is a <code><b>
   * de.renew.util.RenewObjectOutputStream</b></code>, its
   * feature of cutting down the recursion depth by delaying
   * the serialization of some fields will be used.
   * </p>
   *
   * Added Apr 18 2000  Michael Duvigneau
   *
   * @param output target stream (see note about RenewObjectOutputStream)
   * @see de.renew.util.RenewObjectOutputStream
   **/
  public static void saveAllNets(ObjectOutput output) 
         throws IOException {

    Enumeration saveList;
    RenewObjectOutputStream rOut = null;
    if (output instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)output;

    if (rOut != null) rOut.beginDomain(Net.class);

    // Save all currently known nets.
    output.writeInt(netsByName.size());
    saveList=netsByName.elements();
    while (saveList.hasMoreElements()) {
      output.writeObject(saveList.nextElement());
    }

    // If a RenewObjectOutputStream is used, write
    // all delayed fields NOW.
    if (rOut != null) rOut.writeDelayedObjects();

    if (rOut != null) rOut.endDomain(Net.class);
  }

  /**
   * Reads Nets stored by <code>saveAllInstances()</code>
   * and adds the new <code>Net</code>s to the list
   * of known nets by name.
   * </p><p>
   * If the given <code>ObjectInput</code> is a <code>
   * de.renew.util.RenewObjectInputStream</code>, the
   * neccessary steps to cover delayed serialization will
   * be made.
   * </p><p>
   * The ObjectInputStream will be read using
   * <code>de.renew.util.ClassSource</code> to provide
   * its ability of reloading all user defined classes.
   * </p>
   *
   * Added Apr 18 2000  Michael Duvigneau
   *
   * @param input source stream (see notes above)
   * @see #saveAllNets
   * @see de.renew.util.ClassSource
   * @see de.renew.util.RenewObjectInputStream
   **/
  public static void loadNets(ObjectInput input) 
         throws IOException, ClassNotFoundException {

    // Read all stored Nets.
    int count = input.readInt();
    Net newNet;
    try {
      for (int i = 0; i < count; i++) {
	newNet=(Net)de.renew.util.ClassSource.readObject(input);
      }
    } catch (ClassCastException e) {
      throw new StreamCorruptedException (
		"Object other than Net found "
		+"when looking for nets: "
		+e.getMessage());
    }
    
    // If a RenewObjectInputStream is used, read
    // all delayed fields NOW.
    if (input instanceof RenewObjectInputStream)
      ((RenewObjectInputStream)input).readDelayedObjects();
  }

}

