package de.renew.simulator;

import de.renew.unify.*;
import collections.*;
import java.io.Serializable;
import java.io.IOException;
import de.renew.util.RenewObjectOutputStream;

public class Place implements Serializable {

  static final long serialVersionUID = -454336128139256244L;

  /**
   * My name.
   **/
  String name;

  /**
   * true, if my instances should output a trace message when they
   * receive an initial token. This is the default.
   **/
  private boolean trace;

  /**
   * My set of token source inscriptions.
   **/
  UpdatableSet inscriptions;

  public Place(Net net, String name) {
    this.name=name;

    trace=true;
    inscriptions=new HashedSet();

    net.add(this);
  }

  public String toString() {
    return name;
  }

  PlaceInstance makeInstance(NetInstance netInstance) 
    throws Impossible
  {
    PlaceInstance instance=new PlaceInstance(netInstance,this);
    
    return instance;
  }

  /**
   * Switch my trace flag on or off.
   *
   * @param trace true, if tracing is desired
   **/
  public void setTrace(boolean trace) {
    this.trace=trace;
  }

  /**
   * Am I being traced?
   *
   * @return true, if my trace flag is set
   **/
  public boolean getTrace() {
    return trace;
  }

  /**
   * Give me another token source.
   **/
  public void add(TokenSource tokenSource) {
    inscriptions.include(tokenSource);
  }
  
  /**
   * Remove one my current token sources.
   **/
  public void remove(TokenSource tokenSource) {
    inscriptions.exclude(tokenSource);
  }


  /**
   * Serialization method, behaves like default writeObject
   * method except using the domain trace feature, if the
   * output is a RenewObjectOutputStream.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) rOut.endDomain(this);
  }


  // ---- ID-Handling ------------------------------------------------

  /**
   * Holds the internal ID assigned to this place.
   * @see #setID
   **/
  private int _id;

  /**
   * Assigns the given internal ID (e.g. the ID of the
   * corresponding shadow net element) to this place.
   * This should be done at creation time by the compiler.
   *
   * @see de.renew.shadow.ShadowNetElement#setID
   **/
  public void setID(int newID) {
    _id = newID;
    //System.out.println("Place \""+name+"\" was assigned to ID "+_id);
  }

  /**
   * Returns the internal ID assigned to this place.
   * @see #setID
   **/
  public int getID() {
    return _id;
  }

}

