package de.renew.simulator;

import de.renew.expression.*;
import collections.*;
import de.renew.event.*;
import de.renew.unify.*;
import de.renew.util.Lock;
import de.renew.util.Null;
import java.util.Enumeration;
import java.util.Vector;
import java.io.Serializable;
import java.io.IOException;
import de.renew.util.DelayedFieldOwner;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

// I will collect tokens for a single place instance. I also
// keep count of the number of tokens that are currently being tested.
//
// Whenever my marking is changed I will notify my triggerables.
//
// Anybody who needs to make sure that my marking does not change should
// lock on me. Synchronisation does not help!

public class PlaceInstance 
    implements PlaceEventProducer, Serializable, DelayedFieldOwner {

  static final long serialVersionUID = -6409146722455348472L;

  private NetInstance netInstance;
  private Place place;
  
  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  transient private TokenBag freeTokens;

  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  transient private TokenBag testedTokens;

  /**
   * Needed for technical reasons to distinguish the 
   * two TokenBag fields on deserialization. See comment
   * in <code>reassignField()</code> for further details.
   **/
  transient private int bagCounter;

  private UpdatableSeq initTokens;

  /** 
   * All elements contained in this index must also
   * appear in <code>testIndex</code>.
   *
   * This field is transient and will be recomputed
   * on deserialization.
   **/
  transient private TupleIndex freeIndex;

  /** 
   * Must contain at least all elements also 
   * contained in <code>freeIndex</code>.
   *
   * This field is transient and will be recomputed
   * on deserialization.
   **/
  transient private TupleIndex testIndex;

  /** 
   * This field is not really transient, but as we want
   * to cut down the serialization recursion depth, it
   * is serialized manually.
   **/
  transient private TriggerableCollection triggerables;

  // lockOrder and lock were originally final, but to allow the
  // creation of new values on deserialization the modifier
  // had to be removed.
  public transient long lockOrder;
  public transient Lock lock;

  /**
   * The triggers that might notify me.
   **/
  private transient UpdatableSet listeners=new HashedSet();

  PlaceInstance(NetInstance netInstance, Place place)
    throws Impossible
  {
    this.netInstance=netInstance;
    this.place=place;

    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();

    freeTokens=new TokenBag();
    testedTokens=new TokenBag();

    freeIndex=new TupleIndex();
    testIndex=new TupleIndex();

    triggerables=new TriggerableCollection();

    // Let's calculate the initial bag of tokens.
    initTokens=new LinkedList();
    CollectionEnumeration inscriptions=place.inscriptions.elements();
    while (inscriptions.hasMoreElements()) {
      VariableMapper mapper=new VariableMapper();
      Variable thisVariable=mapper.map(new LocalVariable("this"));
      Unify.unify(thisVariable,netInstance,null);

      TokenSource tokenSource=(TokenSource)inscriptions.nextElement();
      Object token=tokenSource.createToken(mapper);

      if (netInstance.getNet().earlyTokens) {
	// I am requested to insert the tokens now.
	// This has the downside that I cannot write
	// the token movements to the log.
	silentlyInsertToken(token);
      } else {
	initTokens.insertFirst(Null.convert(token));
      }
    }
  }

  public String toString() {
    return netInstance.toString()+"."+place.toString();
  }

  public NetInstance getNetInstance() {
    return netInstance;
  }

  public Place getPlace() {
    return place;
  }

  // Access the triggerables. We do not lock the place instance during
  // accesses to the triggerables contrary to earlier implementations.
  // The triggerables have to take care of locking themselves.
  public TriggerableCollection triggerables() {
    return triggerables;
  }

  public void addPlaceEventListener(PlaceEventListener 
    listener)
  {
    synchronized(listeners) {
      listeners.include(listener);
    }
  }
  
  public void removePlaceEventListener(PlaceEventListener
    listener)
  {
    synchronized(listeners) {
      listeners.exclude(listener);
    }
  }

  public CollectionEnumeration getDistinctTokens() {
    return freeIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTokens(Object pattern) {
    return freeIndex.getPossibleMatches(pattern);
  }

  public CollectionEnumeration getDistinctTestableTokens() {
    return testIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTestableTokens(Object pattern) {
    return testIndex.getPossibleMatches(pattern);
  }

  public int getNumberOfTokens() {
    return freeTokens.getSize();
  }

  public int getNumberOfTestedTokens() {
    // We do not count multiple tests on a single token as
    // multiple tokens.
    return testedTokens.getUniqueSize();
  }

  public int getNumberOfTestableTokens() {
    return getNumberOfTokens()+getNumberOfTestedTokens();
  }

  boolean containsToken(Object token) {
    return freeTokens.includes(token);
  }

  public int getTokenCount(Object token) {
    return freeTokens.getMultiplicity(token);
  }

  // Here consider only already tested tokens.
  public boolean containsTestedToken(Object token) {
    return testedTokens.includes(token);
  }
  // Here we consider both kinds of tokens.
  boolean containsTestableToken(Object token) {
    // nulls will be converted by the called methods.
    return containsToken(token) || containsTestedToken(token);
  }

  void removeToken(Object token) {
    freeTokens.removeOneOf(token);

    if (!freeTokens.includes(token)) {
      freeIndex.remove(token);
      if (!testedTokens.includes(token)) {
	testIndex.remove(token);
      }
    }

    triggerables.proposeSearch(false,true);

    synchronized(listeners) {
      TokenEvent te=new TokenEvent(this,token);
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenRemoved(te);
      }
    }
  }

  // Multiple tests on the same token do not remove
  // the token multiple times. Instead, the token is
  // removed once, and the number of removals is recorded
  // by a multiple insertion into the bag of tested tokens.
  void testToken(Object token) {
    boolean wantNotify=false;

    if (!testedTokens.includes(token)) {
      freeTokens.removeOneOf(token);
      if (!freeTokens.includes(token)) {
	freeIndex.remove(token);
      }

      // We must notify the listeners only if a free token has
      // become tested. If a token was already tested, the new
      // test simply delays the time span until a possible
      // activation.
      triggerables.proposeSearch(false,true);

      wantNotify=true;
    }

    testedTokens.add(token);

    if (wantNotify) {
      synchronized(listeners) {
	TokenEvent te=new TokenEvent(this,token);
	CollectionEnumeration enum=listeners.elements();
	while (enum.hasMoreElements()) {
	  PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	  listener.tokenTested(te);
	}
      }
    }
  }

  // This method must be used with care. It inserts tokens
  // into the place without notification, although it updates
  // the indices.
  public void silentlyInsertToken(Object token) {
    if (!freeTokens.includes(token)) {
      freeIndex.insert(token);
      // Maybe there is an appropriate token being tested?
      if (!testedTokens.includes(token)) {
	// No, remember the new token.
	testIndex.insert(token);
      }
    }

    freeTokens.add(token);
  }

  void insertToken(Object token) {
    silentlyInsertToken(token);
    triggerables.proposeSearch(true,false);

    synchronized(listeners) {
      TokenEvent te=new TokenEvent(this,token);
      CollectionEnumeration enum=listeners.elements();
      while (enum.hasMoreElements()) {
	PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	listener.tokenAdded(te);
      }
    }
  }

  // Undo a testToken call. Once the last token is returned,
  // a token is reinserted into the bag of free tokens.
  void untestToken(Object token) {
    testedTokens.removeOneOf(token);
    // No need to change the test index, because a tested token
    // will become free or an equivalent token will remain tested.

    if (!testedTokens.includes(token)) {
      if (!freeTokens.includes(token)) {
	freeIndex.insert(token);
      }
      freeTokens.add(token);

      // We must notify the listeners only if a tested token has
      // become free. If a token is still tested, the 
      // untest simply shortens the time span until a possible
      // activation.
      triggerables.proposeSearch(true,false);

      synchronized(listeners) {
	TokenEvent te=new TokenEvent(this,token);
	CollectionEnumeration enum=listeners.elements();
	while (enum.hasMoreElements()) {
	  PlaceEventListener listener=(PlaceEventListener)enum.nextElement();
	  listener.tokenUntested(te);
	}
      }
    }
  }

  /**
   * My net will call this method if its creation is confirmed.
   * I will free all tokens that I kept during my initialisation.
   * You must call this method at most once.
   **/
  void createConfirmation(Tracer tracer) {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The listeners must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    synchronized(listeners) {
      CollectionEnumeration enum=initTokens.elements();
      // Anything to do?
      if (enum.hasMoreElements()) {
	// Yes, move all tokens to their destination.
	do {
	  Object token=Null.convert(enum.nextElement());
	  if (place.getTrace()) {
            tracer.trace("    Putting "+token+
              " into "+this);
	  }
	  silentlyInsertToken(token);
	} while (enum.hasMoreElements());
	
	triggerables.proposeSearch(true,false);

	// Just notify the listeners once.
	PlaceEvent pe=new PlaceEvent(this);
	CollectionEnumeration listenerEnum=listeners.elements();
	while (listenerEnum.hasMoreElements()) {
	  PlaceEventListener listener=
	    (PlaceEventListener)listenerEnum.nextElement();
	  listener.markingChanged(pe);
	}
      }
      
      // Null the field to allow garbage collection.
      initTokens=null;
    }
  }

  /**
   * Serialization method, behaves like default writeObject
   * method except storing the not-really-transient fields
   * freeTokens, testedTokens and triggerables.
   * If the stream used is a RenewObjectOutputStream, these
   * fields are delayed to cut down recursion depth.
   * The domain trace feature of this special stream is also
   * used.
   * @see de.renew.util.RenewObjectOutputStream
   **/
  private void writeObject(java.io.ObjectOutputStream out) throws IOException {
    RenewObjectOutputStream rOut = null;
    if (out instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)out;
    if (rOut != null) rOut.beginDomain(this);
    out.defaultWriteObject();
    if (rOut != null) {
      rOut.delayedWriteObject(freeTokens, this);
      rOut.delayedWriteObject(testedTokens, this);
      rOut.delayedWriteObject(triggerables, this);
      rOut.endDomain(this);
    } else {
      out.writeObject(freeTokens);
      out.writeObject(testedTokens);
      out.writeObject(triggerables);
    }
  }

  /**
   * Deserialization method, behaves like default readObject
   * method except restoring additional transient fields.
   * Creates new lock object and gets a new lock order ticket.
   * The method also restores the not-really-transient fields
   * freeTokens, testedTokens and triggerables, <b>if</b> the
   * stream used is <b>not</b> a RenewObjectInputStream.
   * In that case it also recomputes the transient fields
   * freeIndex and testedIndex.
   **/
  private void readObject(java.io.ObjectInputStream in)
               throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();
    listeners=new HashedSet();

    if (in instanceof RenewObjectInputStream) {
      // Do nothing, the fields will be
      // reassigned by the stream soon.
      // Set bagCounter to 1:
      // See comment in reassignField()
      // for reason to do this.
      bagCounter = 1;
    } else {
      freeTokens = (TokenBag)in.readObject();
      testedTokens = (TokenBag)in.readObject();
      triggerables = (TriggerableCollection)in.readObject();
      recomputeIndexes();
    }
  }

  /** 
   * Method used on deserialization by RenewObjectInputStream.
   * Reassigns values to the not-really-transient fields
   * freeTokens, testedTokens and triggerables, one at a
   * time.
   * Also recomputes freeIndex and testIndex when the
   * corresponding TokenBag gets reassigned.
   **/
  public void reassignField(Object value) throws java.io.IOException {

    if (value instanceof TokenBag) {
      // On deserialization we expect to get two delayed instances
      // of TokenBags. They can only be distinguished by the
      // order of their assignment.
      // In the readObject() method the counter was initialized to 1
      // now we can count and assign them to the correct fields.
      TokenBag aBag = (TokenBag)value;
      switch (bagCounter) {
      case 1:
	  bagCounter++;
	  freeTokens = aBag;
	  break;
      case 2:
	  bagCounter = -1;
	  testedTokens = aBag;
	  recomputeIndexes();
	  break;
      default:
	  // That's no good!
	  throw new java.io.NotSerializableException(
		    "Too many TokenBags or "
		    +"TokenBag unexpectedly given to reassign().");
      }
    } else if (value instanceof TriggerableCollection) {
      triggerables = (TriggerableCollection)value;
    } else {
	throw new java.io.NotSerializableException(
		  "Value of unexpected type given to reassign().");
    }
  }

  /**
   * Rebuilds both TupleIndex fields (freeIndex and 
   * testIndex) from scratch, using the elements of
   * the TokenBag fields freeTokens and testedTokens.
   * Needed by the deserialization methods readObject()
   * and reassignField() to update the transient
   * TupleIndex fields.
   **/
  private void recomputeIndexes() {
      freeIndex=new TupleIndex();
      addToIndex(freeIndex,freeTokens);
      testIndex=new TupleIndex();
      addToIndex(testIndex,freeTokens);
      addToIndex(testIndex,testedTokens);
  }

  /**
   * Inserts all elements of the given TokenBag into
   * the given TupleIndex. Used by recomputeIndexes().
   **/
  private void addToIndex(TupleIndex index, TokenBag tokens) {
      Enumeration tokenList = tokens.uniqueElements();

      while (tokenList.hasMoreElements()) {
	  Object token=tokenList.nextElement();
	  index.insert(token);
      }
  }
    
}

