package de.renew.simulator;

import collections.*;

import java.io.ObjectOutput;
import java.io.ObjectInput;
import java.io.IOException;
import java.io.StreamCorruptedException;
import de.renew.util.RenewObjectOutputStream;
import de.renew.util.RenewObjectInputStream;

/**
 * The class search queue keeps track of the transitions
 * instances that must be searched for activated bindings.
 * This is done statically, so that a net instance
 * can register its transitions as possibly activated even if it
 * does not know of any specific searcher.
 * A searcher can ask the search queue for
 * possibly activated transitions.
 *
 * In theory, this might also be incorporated in the Searcher
 * class, but someday there might be multiple searchers
 * acting concurrently and it would be quite confusing if the
 * different searchers were synchronized via a static search queue
 * inside the searcher class. By separating this class,
 * it gets clear that a single common search queue has been implemented.
 * In the sense of load balancing it is also sensible to have a central
 * search queue.
 *
 * @author Olaf Kummer
 **/

public abstract class SearchQueue {
  private static SearchQueueData instance=new RandomSearchQueue();
  private static UpdatableSeq listeners=new LinkedList();

  // Set the instance of SearchQueueData that manages the
  // queue policy. This enables us to change search queue
  // policies at runtime under the control of the simulator.
  public synchronized static void setQueue(SearchQueueData queue) {
    queue.init(instance.elements());
    instance=queue;
  }

  // Clear all pending enabled transition instances in the queue.
  // This is required when a new simulation is started. Note that
  // there might be threads running in the background that
  // might reintroduce transition instances into the queue.
  // So be sure to kill everything in sight before calling
  // this method.
  public synchronized static void reset() {
    try {
      // Try to create search queue of the same type first.
      instance=(SearchQueueData)instance.getClass().newInstance();
    }
    catch (Exception e) {
      // Create a default search queue.
      instance=new RandomSearchQueue();
    }
  }

  public synchronized static void insertListener(
    SearchQueueListener listener)
  {
    listeners.insertFirst(listener);
  }
  
  protected synchronized static void notifyListeners() {
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      ((SearchQueueListener)enum.nextElement()).searchQueueNonempty();
    }
    listeners.clear();
  }

  // Forward calls to the instance. This is required
  // so that we can already obtain a lock that will allow
  // a clean switching of instances.
  public synchronized static void include(
    Searchable searchable)
  {
    // Notify a searchable that its enabledness will be detected
    // by a future search. This has to be done regardless
    // of the policy of the search queue.
    searchable.triggers().clear();

    // Do it.
    instance.include(searchable);

    // Does anybody want a searchable?
    notifyListeners();
  }

  public synchronized static boolean isEmpty() {
    return instance.isEmpty();
  }

  public synchronized static Searchable extract() {
    return instance.extract();
  }

  /**
   * Writes all entries currently in the queue
   * (and all their associated field data)
   * to the given stream. The written information
   * should describe the complete current simulation
   * state.
   * But this information does not neccessarily include all
   * nets possibly required by future simulation steps.
   * <p>
   * If the given <code>ObjectOutput</code> is a <code><b>
   * de.renew.util.RenewObjectOutputStream</b></code>, its
   * feature of cutting down the recursion depth by delaying
   * the serialization of some fields will be used.
   * </p><p>
   * <b>Caution:</b> In order to get consistent data written
   * to the stream, you have to ensure that there are no
   * concurrent modifications of the simulation state.
   * This method is not able to lock the simulation.
   * </p>
   *
   * Added Feb 29 2000  Michael Duvigneau
   *
   * @param output target stream (see note about RenewObjectOutputStream)
   * @see de.renew.util.RenewObjectOutputStream
   **/
  public synchronized static void saveQueue(ObjectOutput output) 
	 throws IOException {

    CollectionEnumeration saveList;
    RenewObjectOutputStream rOut = null;
    if (output instanceof RenewObjectOutputStream)
      rOut = (RenewObjectOutputStream)output;

    if (rOut != null) rOut.beginDomain(SearchQueue.class);

    // Save all entries currently in the Queue.
    output.writeInt(instance.size());
    saveList = instance.elements();
    while (saveList.hasMoreElements()) {
      output.writeObject(saveList.nextElement());
    }

    // If a RenewObjectOutputStream is used, write
    // all delayed fields NOW.
    if (rOut != null) rOut.writeDelayedObjects();

    if (rOut != null) rOut.endDomain(SearchQueue.class);
  }


  /**
   * Restores queue elements saved by <code>saveQueue()</code>.
   * Adds all stored elements to the queue.
   * <p>
   * If the given <code>ObjectInput</code> is a <code>
   * de.renew.util.RenewObjectInputStream</code>, the
   * neccessary steps to cover delayed serialization will
   * be made.
   * </p><p>
   * The object input stream will be read using
   * <code>de.renew.util.ClassSource</code> to provide
   * its ability of reloading all user defined classes.
   * </p>
   *
   * Added Apr 11 2000  Michael Duvigneau
   *
   * @param input source stream (see note about RenewObjectInputStream above)
   * @see de.renew.util.ClassSource
   * @see de.renew.util.RenewObjectInputStream
   **/
  public synchronized static void loadQueue(ObjectInput input) 
	 throws IOException, ClassNotFoundException {

    // Read all entries.
    int count = input.readInt();
    Searchable[] readList = new Searchable[count]; 
    try {
      for (int i = 0; i < count; i++) {
	readList[i]=(Searchable)de.renew.util.ClassSource.readObject(input);
      }
    } catch (ClassCastException e) {
      throw new StreamCorruptedException (
		"Object other than Searchable found "
		+"when looking for SearchQueue elements: "
		+e.getMessage());
    }

    // If a RenewObjectInputStream is used, read
    // all delayed fields NOW.
    if (input instanceof RenewObjectInputStream)
      ((RenewObjectInputStream)input).readDelayedObjects();

    // Insert all entries into the queue.
    for (int i = 0; i < count; i++) {
      include(readList[i]);
    }

  }
}
