package de.renew.simulator;

import collections.*;
import java.util.Vector;
import java.util.Enumeration;
import de.renew.util.Semaphor;
import de.renew.unify.*;

public class Searcher {
  // A searcher keeps track of various possibilities to
  // refine a partially known binding. By default the
  // possibility with the least number of branches to examine
  // is chosen, but this construction allows a simple exchange
  // of algorithms.
  //
  // The algorithm may proceed to bind tokens to arc expressions,
  // downlinks to individual transitions, or
  // transitions to arc expressions.

  private boolean bindingPossible;
  private Finder finder;

  private UpdatableSet binders;
    // UpdatableSet<Binder>
  private UpdatableSet occurrences;
    // UpdatableSet<Occurrence>
  private Triggerable primaryTriggerable;

  public CalculationChecker calcChecker;
  public StateRecorder recorder;
  public TokenReserver reserver;

  public Searcher() {
    binders=new HashedSet();
    occurrences=new HashedSet();

    calcChecker=new CalculationChecker();
    recorder=new StateRecorder();
    reserver=new TokenReserver();
  }

  // Returns true, if we should abort the current
  // search for new bindings. This method may only be called
  // by myself or by invoked binders during the search.
  public boolean isCompleted() {
    boolean result=finder.isCompleted();
    if (result) {
      // Leave a note that a binding was possible.
      // We are aborting now, so we cannot determine whether
      // there is indeed a binding.
      bindingPossible=true;
    }
    return result;
  }

  private void setPrimaryTriggerable(Triggerable triggerable) {
    // Clear all triggers previously associated to the triggerable.
    if (triggerable!=null) {
      triggerable.triggers().clear();
    }

    primaryTriggerable=triggerable;
  }

  public void insertTriggerable(TriggerableCollection triggerables) {
    // If there is a primary triggerable, we register it
    // at the given collection of triggerables.
    if (primaryTriggerable!=null) {
      triggerables.include(primaryTriggerable);
    }
  }

  // Maintain a collection of transition occurrences.
  public void addOccurrence(Occurrence occurrence) {
    occurrences.include(occurrence);
  }

  public void removeOccurrence(Occurrence occurrence) {
    occurrences.exclude(occurrence);
  }

  public CollectionEnumeration getOccurrences() {
    return occurrences.elements();
  }

  // A convenience method that allows the simple generation
  // of all executables that are associated to the occurrences.
  // 
  // Unlike earlier version of this method, the returned
  // executables are unsorted!
  public CollectionEnumeration makeExecutables() {
    LinkedList executables=new LinkedList();
    
    Copier copier=new Copier();

    Enumeration transitionOccurrences=
      getOccurrences();
    while (transitionOccurrences.hasMoreElements()) {
      Occurrence occurrence=
	(Occurrence)transitionOccurrences.nextElement();
      executables.prependElements(occurrence.makeExecutables(copier));
    }

    return executables.elements();
  }

  // Find the binder with the minimum binding badness
  // Return null, if no binder wants to try.
  private Binder selectBestBinder() {
    Binder bestBinder=null;
    int bestBadness=BindingBadness.max;
    Enumeration enum=binders.elements();
    while (enum.hasMoreElements() && !isCompleted()) {
      Binder binder=(Binder)enum.nextElement();
      int badness=binder.bindingBadness(this);
      if (badness<bestBadness) {
	bestBinder=binder;
	bestBadness=badness;
      }
    }
    return bestBinder;
  }

  public void search() {
    if (!binders.isEmpty()) {
      Binder binder=selectBestBinder();
      if (binder!=null && !isCompleted()) {
	int checkpoint=recorder.checkpoint();
	removeBinder(binder);
        binder.bind(this);
	addBinder(binder);
	recorder.restore(checkpoint);
      }
    } else {
      if (calcChecker.isConsistent()) {
	bindingPossible=true;
	finder.found(this);
      }
    }
  }

  public void addBinder(Binder binder) {
    binders.include(binder);
  }
  public void removeBinder(Binder binder) {
    binders.exclude(binder);
  }
  public void addBinders(Enumeration newBinders) {
    binders.includeElements(newBinders);
  }
  public void removeBinders(Enumeration oldBinders) {
    binders.removeElements(oldBinders);
  }

  private void startSearch() {
    // Make sure that the current state recorder does not waste
    // space with now obsolete bindings.
    recorder.restore();

    // Ensure that the temporary containers are initially empty.
    // If they are not, something went wrong previously,
    // so we report en error.
    if (!occurrences.isEmpty() || !binders.isEmpty()) {
      throw new RuntimeException("Searcher was not in idle state "+
	"at the start of a search.");
    }
  }

  private void searchAndRecover(Searchable searchable,
    Triggerable triggerable)
  {
    setPrimaryTriggerable(triggerable);

    try {
      // Is this searchable object activated?
      searchable.startSearch(this);
    }
    catch (Exception e) {
      e.printStackTrace();
      // We reinsert this searchable only when a binding was found,
      // because otherwise there is no real chance that we
      // could have more luck next time.
    }
    
    // Reset state to allow garbage collection.
    binders.clear();
    occurrences.clear();
    
    calcChecker.reset();
    recorder.restore();
    reserver.reset();

    setPrimaryTriggerable(null);
  }

  public void searchLoop(Finder finder) {
    // Prepare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // Make a list of all the transition instances
    // that could fire, even if they were rejected by the
    // finder.
    UpdatableSeq wasActivated=new LinkedList();

    // Try all transitions that might be enabled.
    // If a successful binding was found (though not
    // neccessarily accepted), requeue this transition
    // at the end of the search.
    // If no bindings are possible, forget about this 
    // transition. A change to a relevant place will
    // trigger its reinsertion.
    while ((!isCompleted()) && (!SearchQueue.isEmpty())) {
      // Take a note that no binding has been found so far.
      bindingPossible=false;

      // Get a searchable.
      Searchable searchable=SearchQueue.extract();

      // Perform the search. The searchable itself is used
      // as the triggerable, because it was taken from the search queue.
      searchAndRecover(searchable,searchable);
      
      // We must check for a reinsertion of this transition
      // into the list of activated elements, regardless whether this
      // transition instance is actually selected by the binder.
      if (bindingPossible) {
	// Ok, reinsert this transition.
	wasActivated.insertFirst(searchable);
      }
    }

    // Reinsert the activated transitions into the search list.
    // The intermediate list ensures the termination of the inner loop,
    // when a binder does not accept any of the proposed bindings.
    CollectionEnumeration enum=wasActivated.elements();
    while (enum.hasMoreElements()) {
      SearchQueue.include((Searchable)enum.nextElement());
    }

    // Forget the finder.
    this.finder=null;
  }

  // Try to find a binding for a searchable. If neccessary,
  // reinsert the searchable into the search queue.
  //
  // The triggerable provided to this method will be notified
  // whenever the searchable might get a different set of
  // activated bindings. If the searchable was extracted from 
  // the search queue, the searchable itself should be used
  // as the triggerable.
  //
  // The handleInsertion flag should be true, if the searchable
  // was drawn from the search queue.
  void searchOnce(Finder finder, Searchable searchable,
    Triggerable triggerable)
  {
    // Perpare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // Take a note that no binding has been found so far.
    bindingPossible=false;
    
    // Perform the search.
    searchAndRecover(searchable,triggerable);
      
    // Forget finder.
    this.finder=null;
  }

  public void initiatedSearch(ChannelTarget channelTarget,
    String name, Variable params, boolean isOptional, Finder finder,
    Triggerable triggerable)
  {
    // Perpare a new search that starts from scratch.
    startSearch();

    // Remember the finder.
    this.finder=finder;

    // Possible register for state change notifications.
    setPrimaryTriggerable(triggerable);

    try {
      // We want to find a transition within the specified
      // net instance that can fire. This method is
      // usually called to generate the very first tokens
      // in a net. In a sense, this resembles the main(args)
      // call of the runtime environment.
      //
      // If the argument isOptional is true, the invoked net
      // need not provide an appropriate channel at all. But if it does,
      // the synchronisation must succeed.
      Variable targetVariable=new Variable(channelTarget,recorder);
      Binder initialBinder=
	new ChannelBinder(targetVariable,name,params,isOptional);
      addBinder(initialBinder);
      search();
      removeBinder(initialBinder);
      recorder.restore();
    }
    catch (Exception e) {
      e.printStackTrace();
      // We reinsert this searchable only when a binding was found,
      // because otherwise there is no real chance that we
      // could have more luck next time.
    }

    // Reset state to allow garbage collection.
    binders.clear();
    occurrences.clear();
    
    calcChecker.reset();
    recorder.restore();
    reserver.reset();

    setPrimaryTriggerable(null);
    finder=null;
  }
}

