package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

public class SimulatorHelper {
  public static NetInstance fireInitially(Net net,
    Tracer tracer, boolean asynchronous)
  {
    // Search for the initial transition and fire it.
    // We make sure that the primary instance exists
    // only after a successful initial synchronisation.
    // We create a searcher especially for this purpose.
    // This is not overly efficient, but we do not expect
    // to fire a primary instance often.
    Searcher searcher=new Searcher();

    // Create the primary instance.
    NetInstance primaryInstance;
    try {
      primaryInstance=net.makeInstance();
    }
    catch (Impossible e) {
e.printStackTrace();
      primaryInstance=null;
    }

    // If the creation of the primary instance succeeded,
    // Try to perform the initial synchronisation.
    ExecuteFinder finder=new ExecuteFinder();
    if (primaryInstance!=null) {
      // Fire the primary transition to get an
      // initial marking. The parameters could be
      // dependent on the parameters of this method.
      // No need to undo this action.
      Variable variable=new Variable(new Tuple(0),null);
      
      // Prepare the execution of the net. If no channel with
      // name "new" is present, simply ignore the synchronisation.
      searcher.initiatedSearch(primaryInstance,
	"new",variable,true,finder,null);

      // Did we find a binding?
      if (!finder.isCompleted()) {
	primaryInstance=null;
      }
    }

    // Execute the binding.
    if (primaryInstance!=null) {
      // Confirm the permanent activation for the
      // fully spontaneous transitions of this net. The
      // confirmation is only done if the synchronisation succeeds,
      // because conceptionally, the net instance is not created
      // otherwise. But the confirmation must be done before the
      // initial transition is executed, because another binding
      // is sought immediately.
      primaryInstance.createConfirmation(tracer);
      
      // Execute the initial transition. Its binding
      // is recorded in the finder.
      finder.execute(tracer,asynchronous);
    }

    return primaryInstance;
  }

  // This method checks a searchable for enabledness.
  // A searcher must be provided for this method. The method
  // uses the given triggerable to store a new set of triggers.
  public static boolean isFirable(Searchable searchable, 
    Triggerable triggerable, Searcher searcher)
  {
    EnablednessFinder finder=new EnablednessFinder();
    searcher.searchOnce(finder,searchable,triggerable);
    return finder.isEnabled();
  } 

  // This method checks a searchable for enabledness. It
  // uses the given triggerable to store a new set of triggers.
  public static boolean isFirable(Searchable searchable, 
    Triggerable triggerable)
  {
    Searcher searcher=new Searcher();
    return isFirable(searchable,triggerable,searcher);
  } 

  // This method fires one binding. It resets the triggers
  // of the transition instance and makes sure the reinsert
  // it into the search queue, if neccessary.
  public static boolean fireOneBinding(TransitionInstance instance, 
    Tracer tracer, boolean asynchronous, Searcher searcher)
  {
    ExecuteFinder finder=new ExecuteFinder();
    searcher.searchOnce(finder,instance,instance);

    if (finder.isCompleted()) {
      // Ok, reinsert this searchable. The finder used in the
      // previous method call cannot be interrupted and hence it
      // is suffiecient to check for executability to determine
      // enabledness.
      SearchQueue.include(instance);

      finder.execute(tracer,asynchronous);
      return true;
    }
    return false;
  } 

  public static boolean fireOneBinding(TransitionInstance instance,
    Tracer tracer, boolean asynchronous)
  {
    Searcher searcher=new Searcher();
    return fireOneBinding(instance,tracer,asynchronous,searcher);
  } 

  // This method collects all bindings for a searchable.
  // A searcher must be provided for this method. The method
  // uses the given triggerable to store a new set of triggers.
  public static CollectionEnumeration findAllBindings(Searchable searchable, 
    Triggerable triggerable, Searcher searcher)
  {
    CollectingFinder finder=new CollectingFinder();
    searcher.searchOnce(finder,searchable,triggerable);
    return finder.bindings();
  } 

  // This method collects all bindings for a searchable. It
  // uses the given triggerable to store a new set of triggers.
  public static CollectionEnumeration findAllBindings(Searchable searchable, 
    Triggerable triggerable)
  {
    Searcher searcher=new Searcher();
    return findAllBindings(searchable,triggerable,searcher);
  } 

  // The lock method must be called first. It cannot fail,
  // but it might take a while.
  private static void lock(final Seq executables) {
    CollectionEnumeration enum=executables.elements();
    while (enum.hasMoreElements()) {
      ((EarlyExecutable)enum.nextElement()).lock();
    }
  }

  private static void unlock(final Seq executables) {
    CollectionEnumeration enum=executables.elements();
    while (enum.hasMoreElements()) {
      ((EarlyExecutable)enum.nextElement()).unlock();
    }
  }

  // The verify method must be called second. If it fails with
  // an exception, the transition cannot fire. If it succeeds,
  // it must be followed by a call to executeEarly().
  private static void verify(final Seq executables) throws Impossible {
    UpdatableSeq verified=new LinkedList();

    // Verify all early executables.
    try {
      CollectionEnumeration enum=executables.elements();
      while (enum.hasMoreElements()) {
        EarlyExecutable executable=(EarlyExecutable)enum.nextElement();
	executable.verify();
	// Ok, I got the verification. Now I must make sure to
	// treat this executable correctly: rollback or execute.
	verified.insertFirst(executable);
      }
    }
    catch (Impossible e) {
      // Something failed. Probably a missing token.
      CollectionEnumeration enum=verified.elements();
      while (enum.hasMoreElements()) {
	((EarlyExecutable)enum.nextElement()).rollback();
      }
      // I undid all actions (in reverse order, if that matters).
      // Now I propagate the orginal exception to my caller.
      throw e;
    }
  }

  private static void executeEarly(final Seq executables, 
    final Tracer tracer)
  {
    // All verifications succeeded. Now we call all early
    // executables in their appropriate order.

    CollectionEnumeration enum=executables.elements();
    while (enum.hasMoreElements()) {
      ((EarlyExecutable)enum.nextElement()).execute(tracer);
    }
  }

  static void executeLate(final Seq executables, 
    final Tracer tracer)
  {
    // All early executables were correctly executed and
    // the locks were released. Now we call all late
    // executables in their appropriate order. We do this
    // in a separate thread to avoid locking up the simulation.
    
    CollectionEnumeration lateEnum=executables.elements();
    while (lateEnum.hasMoreElements()) {
      LateExecutable executable=(LateExecutable)lateEnum.nextElement();
      if (executable.isLong()) {
	de.renew.util.Detacher.detach();
      }
      try {
	executable.execute(tracer);
      }
      catch (Exception e) {
	tracer.trace("Exception occurred during late phase of firing:");
	tracer.trace(e.toString());
	e.printStackTrace();
	// Do not proceed. Some variables might have remained
	// unbound, so it is unlikely that the following
	// executables will succeed.
	tracer.complete();
	return;
      }
    }
    // Everything is finished. Let me just inform the tracer.
    tracer.complete();
  }

  // Execute a given sequence of executables.
  public static boolean execute(final CollectionEnumeration executables, 
    final Tracer tracer, final boolean asynchronous)
  {
    LinkedList earlyExecutables=new LinkedList();
    LinkedList lateExecutables=new LinkedList();

    while (executables.hasMoreElements()) {
      Object executable=executables.nextElement();
      if (executable instanceof EarlyExecutable) {
	earlyExecutables.insertFirst(executable);
      } else if (executable instanceof LateExecutable) {
	lateExecutables.insertFirst(executable);
      } else {
	throw new RuntimeException("Unknown type of executable detected: "+
	  executable);
      }
    }

    // Lock in the required order.
    earlyExecutables.sort(new LockComparator());
    lock(earlyExecutables);
	
    try {
      try {
	// Execute in the required order.
	earlyExecutables.sort(new PhaseComparator());
	verify(earlyExecutables);
      
	// Ok, no exception was thrown.
	executeEarly(earlyExecutables,tracer);
      }
      finally {
	// We must make sure to undo every lock even if
	// the execution of some operations failed.
	// The order of the unlock operations does not matter.
	unlock(earlyExecutables);
      }

      // The remaining steps can be done in the background,
      // if we want to do so.
      lateExecutables.sort(new PhaseComparator());
      if (asynchronous) {
	de.renew.util.Detacher.startAttached(new Thread(
          new LateRunnable(lateExecutables,tracer)));
      } else {
	executeLate(lateExecutables,tracer);
      }
      
      return true;
    }
    catch (Impossible e) {
      // The transition could not fire. Probably
      // a token disappeared in the mean time.
      // System.out.println("The calculated binding was disabled.");
      // System.out.println("Searching for an alternative binding.");

      return false;
    }
  }
}
